#!/usr/bin/env python3
#
# Copyright 2017-2022 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
# importlib.resources.files only available since Python>3.9,
# temporarily use backport
from importlib_resources import files

subset_files = [resource.name for resource 
                in files("gfsubsets.data").iterdir()
                if resource.name.endswith("_unique-glyphs.nam")]

SUBSETS = sorted(
    filename.replace("_unique-glyphs.nam", "")
    for filename in subset_files
)
